@@ -3,8 +3,8 @@
 	example.
 
  Copyright (c) 2007 by Educational Technology Resources, Inc.  Others 
- may also hold copyrights on code in this file.  See the CREDITS file in
- the top directory of the distribution for details.
+ may also hold copyrights on code in this file.  See the CREDITS.txt
+ file in the top directory of the distribution for details.
 
  This file is part of MySQL++.
 
@@ -156,9 +156,8 @@ CExampleDlg::OnBnClickedConnectButton()
 
 	if (res) {
 		// Display the result set
-		mysqlpp::Row row;
-		for (mysqlpp::Row::size_type i = 0; row = res.at(i); ++i) {
-			if (ToUCS2(awcTempBuf, kTempBufSize, row.at(0))) {
+		for (size_t i = 0; i < res.num_rows(); ++i) {
+			if (ToUCS2(awcTempBuf, kTempBufSize, res[i][0])) {
 				AddMessage(awcTempBuf);
 			}
 		}
